use crate::vector::Vector;
use rug::{ops::Pow, Integer};
use std::{
    fmt,
    ops::{Index, IndexMut},
};

#[derive(Debug, PartialEq, Clone)]
pub struct Matrix<T> {
    pub n: usize, // number of columns
    pub m: usize, // number of rows
    pub columns: Vec<Vector<T>>,
}

impl<T> Matrix<T> {
    pub fn new(n: usize, m: usize, values: Vec<T>) -> Option<Self> {
        if n * m != values.len() {
            return None;
        }
        // avoid requiring Vec<T>: Clone by building manually
        let mut columns = Vec::with_capacity(n);
        for _ in 0..n {
            columns.push(Vec::with_capacity(m));
        }
        for (i, value) in values.into_iter().enumerate() {
            let col = i % n;
            columns[col].push(value);
        }
        Some(Matrix {
            n,
            m,
            columns: columns.into_iter().map(Vector::from_vec).collect(),
        })
    }
}

impl<T> Matrix<T>
where
    T: Clone + Default + From<Integer>,
{
    pub fn new_lattice(noise_bits: usize, ciphertexts: Vec<T>) -> Option<Self> {
        let n = ciphertexts.len();
        let mut columns = vec![Vec::with_capacity(n); n];

        // First row: [2^(noise+1), ciphertexts[1], ciphertexts[2], ...]
        let two_pow = Integer::from(2u64).pow((noise_bits + 1) as u32);
        columns[0].push(T::from(two_pow));
        for i in 1..n {
            columns[i].push(ciphertexts[i].clone());
        }

        // Subsequent rows form identity matrix with ciphertexts[0]
        for i in 1..n {
            for (j, column) in columns.iter_mut().enumerate().take(n) {
                column.push(if i == j {
                    ciphertexts[0].clone()
                } else {
                    T::default()
                });
            }
        }

        Some(Matrix {
            n,
            m: n,
            columns: columns.into_iter().map(Vector::from_vec).collect(),
        })
    }
}

impl<T> Index<(usize, usize)> for Matrix<T> {
    type Output = T;
    fn index(&self, index: (usize, usize)) -> &Self::Output {
        let (col, row) = index;
        if row >= self.n || col >= self.m {
            panic!("Matrix index out of bounds");
        }
        &self.columns[row][col]
    }
}

impl<T> IndexMut<(usize, usize)> for Matrix<T> {
    fn index_mut(&mut self, index: (usize, usize)) -> &mut Self::Output {
        let (col, row) = index;
        if row >= self.n || col >= self.m {
            panic!("Matrix index out of bounds");
        }
        &mut self.columns[row][col]
    }
}

#[cfg(test)]
mod tests {
    use super::*;
    use rug::Rational;

    macro_rules! int {
        ($x:expr) => {
            Integer::from($x)
        };
    }

    macro_rules! rational {
        ($x:expr) => {
            Rational::from(Integer::from($x))
        };
    }

    #[test]
    fn test_integer_matrix() {
        let m = Matrix::new(2, 2, vec![int!(1), int!(2), int!(3), int!(4)]).unwrap();
        assert_eq!(m[(0, 0)], int!(1));
        assert_eq!(m[(0, 1)], int!(2));
        assert_eq!(m[(1, 0)], int!(3));
        assert_eq!(m[(1, 1)], int!(4));
    }

    #[test]
    fn test_rational_matrix() {
        let m = Matrix::new(
            2,
            2,
            vec![rational!(1), rational!(2), rational!(3), rational!(4)],
        )
        .unwrap();
        assert_eq!(m[(0, 0)], rational!(1));
        assert_eq!(m[(0, 1)], rational!(2));
        assert_eq!(m[(1, 0)], rational!(3));
        assert_eq!(m[(1, 1)], rational!(4));
    }

    #[test]
    fn test_lattice_matrix() {
        let ciphertexts = vec![int!(5), int!(8), int!(12)];
        let noise_bits = 2;
        let lattice = Matrix::new_lattice(noise_bits, ciphertexts).unwrap();

        let expected = vec![
            int!(8),
            int!(8),
            int!(12),
            int!(0),
            int!(5),
            int!(0),
            int!(0),
            int!(0),
            int!(5),
        ];
        let mut actual = Vec::new();
        for col in 0..3 {
            for row in 0..3 {
                actual.push(lattice[(col, row)].clone());
            }
        }
        assert_eq!(actual, expected);
    }

    #[test]
    fn test_rational_lattice() {
        let ciphertexts = vec![rational!(5), rational!(8), rational!(12)];
        let noise_bits = 2;
        let lattice = Matrix::new_lattice(noise_bits, ciphertexts).unwrap();

        let two_pow = Rational::from(Integer::from(2u64).pow(3));
        assert_eq!(lattice[(0, 0)], two_pow);
        assert_eq!(lattice[(0, 1)], rational!(8));
        assert_eq!(lattice[(0, 2)], rational!(12));
    }
}
